package hirondelle.web4j.readconfig;

import java.util.logging.Logger;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;

import hirondelle.web4j.util.Util;

/**
 (UNPUBLISHED) Encapsulate items related to configuration settings.
 
 <P>This class can be used for both <tt>&lt;init-param&gt;</tt> and <tt>&lt;context-param&gt;</tt> 
 items in <tt>web.xml</tt>. 
 
 <P>These items are immutable, and have slightly unusual behaviour. The name and default 
 value are known at compile-time, but the configured value is not known until startup.
 <em>To preserve immutabililty, the {@link #fetch} methods return a new object, 
 instead of changing the state of an existing object.</em> It is this method which actually 
 'reads in' the configured value.
*/
public final class InitParam {
  
  /**
   Construct with a default value. 
    
   @param aName name of the parameter
   @param aDefaultValue default value of the parameter
  */
  public InitParam(String aName, String aDefaultValue){
    this(aName, null, aDefaultValue);    
  }
  
  /** Construct with no default value.  */
  public InitParam(String aName){
    this(aName, null, null);    
  }
  
  /** Return the parameter name passed to the constructor.  */
  public String getName(){
    return fName;
  }
  
  /**
   Return the possibly-null parameter value.
   
   <P>If called immediately after calling a constructor, this method will return null.
   When called after {@link #fetch}, this method will return the configured 
   parameter value (which may in turn be null). Any non-null value is trimmed.
  */
  public String getValue(){
    return  Util.trimPossiblyNull(fValue);
  }
  
  /** 
   Return the default value passed to the constructor.
   
   <P>Any non-null value is trimmed.  
  */
  public String getDefaultValue(){
    return Util.trimPossiblyNull(fDefaultValue);
  }
  
  /**
   Return a new {@link InitParam}, whose value is set by <tt>aConfig</tt>.
   
   <P>The default value of this <tt>InitParam</tt> is taken as the value only if 
   <tt>aConfig</tt> is <tt>null</tt>, or if <tt>aConfig.getInitParameter()</tt> 
   returns <tt>null</tt> for the given parameter value.
   
   <P>This method allows an application to define an <tt>init-param</tt> that takes a 
   default value when it is <em>not</em> specified in <tt>web.xml</tt>. This 
   method can also be used if the init-param has no default value.
   
   <P>This method returns an {@link InitParam} in order to keep this class immutable.
   <P>This method can only be called once.
  */
  public InitParam fetch(ServletConfig aConfig){
    if( Util.textHasContent(fValue) ) {
      throw new IllegalStateException("For InitParam, 'fetch' must be called only once.");
    }
    String paramValue = null;
     if ( aConfig != null ){
       String configValue = aConfig.getInitParameter(fName);
       if ( configValue != null  ){
         //item present in web.xml (but possibly empty)
         paramValue = configValue; 
       }
       else {
         fLogger.fine(
           "Init-param " + Util.quote(fName) + " not in "+ 
           "web.xml. Using default  : " +  Util.quote(fDefaultValue)
         );
         paramValue = fDefaultValue; //possibly null as well
       }
     }
     else {
       //intended for testing environments only, where config object is absent
       paramValue = fDefaultValue; //possibly null as well
     }
     return new InitParam(fName, paramValue, fDefaultValue);
   }
  
  /**
   As in {@link #fetch(ServletConfig)}, but for the rarer occasions when the config is attached to 
   a <tt>ServletContext</tt> (which is used for the web application as a whole, as opposed to the Controller servlet).
   
   <P>The name of this method is distinct from {@link #fetch(ServletConfig)}, to make it difficult to call 
   the wrong method.
  */
  public InitParam fetchFromContext(ServletContext aContext){
    if( Util.textHasContent(fValue) ) {
      throw new IllegalStateException("For InitParam, 'fetch' must be called only once.");
    }
    String paramValue = null;
     if ( aContext != null ){
       String configValue = aContext.getInitParameter(fName);
       if ( configValue != null  ){
         //item present in web.xml (but possibly empty)
         paramValue = configValue; 
       }
       else {
         fLogger.fine(
            "Context-param " + Util.quote(fName) + " not in "+ 
            "web.xml. Using default  : " +  Util.quote(fDefaultValue)
          );
         paramValue = fDefaultValue; //possibly null as well
       }
     }
     else {
       //intended for testing environments only, where config object is absent
       paramValue = fDefaultValue; //possibly null as well
     }
     return new InitParam(fName, paramValue, fDefaultValue);
  }
  
  /** Intended for debugging only.  */
  @Override public String toString(){
    return "Param Name : " + Util.quote(fName) + " Value : " + Util.quote(fValue);
  }
  
  // PRIVATE // 
  private final String fName;
  private final String fValue;
  private final String fDefaultValue;
  private static final Logger fLogger = Util.getLogger(InitParam.class);
  
  private InitParam(String aName, String aValue, String aDefaultValue){
    fName = aName;
    fValue = aValue;
    fDefaultValue = aDefaultValue;
  }
}
