package hirondelle.web4j.security;

import static hirondelle.web4j.util.Consts.NOT_FOUND;
import hirondelle.web4j.action.Operation;
import hirondelle.web4j.request.RequestParser;
import hirondelle.web4j.util.Util;
import hirondelle.web4j.util.WebUtil;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.logging.Logger;
import javax.servlet.ServletConfig;

/**
 Default implementation of {@link UntrustedProxyForUserId}.

 <P>This implementation depends on settings in <tt>web.xml</tt>, which are read in by {@link #init(ServletConfig)}.
  Later, each request URL is parsed by {@link #usesUntrustedIdentifier(RequestParser)}, 
  and an attempt is made to find a match to the aforementioned settings in <tt>web.xml</tt>.
*/
public final class UntrustedProxyForUserIdImpl implements UntrustedProxyForUserId {
  
  /**
    Read in the values of an optional <tt>init-param</tt> in <tt>web.xml</tt> named <tt>UntrustedProxyForUserId</tt>.
    
  <P>This class uses settings in <tt>web.xml</tt> to define requests having ownership constraints that use an untrusted proxy 
  for the user id. It uses a <i>roughly</i> similar style as used for role-based constraints.
  Here is an example of a number of several such ownership constraints defined in <tt>web.xml</tt>:<PRE>&lt;init-param&gt;
  &lt;description&gt;
    Operations having an ownership constraint that uses an untrusted identifier. 
  &lt;/description&gt;
  &lt;param-name&gt;UntrustedProxyForUserId&lt;/param-name&gt;
  &lt;param-value&gt;
    FoodAction.*
    VacationAction.add
    VacationAction.delete
  &lt;/param-value&gt;
&lt;/init-param&gt;
</PRE>

  <P>Each line is treated as a separate constraint, one per line. You can define as many as required.
  The period character separates the 'noun' (the Action) from the 'verb' (the {@link Operation}).
 
  <P>The special '*' character refers to all verbs/operations attached to a given noun/action.
  */
  public static void init(ServletConfig aConfig){
    String rawValue = aConfig.getInitParameter(INIT_PARAM_NAME);
    if( Util.textHasContent(rawValue) ) {
      parseSettings(rawValue);
      fLogger.fine(Util.logOnePerLine(fRestrictedOperations));
    }
    else {
      fLogger.config("No ownership constraints have been set in web.xml. No init-param named " + Util.quote(INIT_PARAM_NAME));
    }
  }

  /** 
    Return <tt>true</tt> only if the given request matches one of the items defined by the <tt>UntrustedProxyForUserId</tt> setting 
    in <tt>web.xml</tt>. 
    
    <P>For example, given the URL : 
    <PRE>'.../VacationAction.list?X=Y'</PRE>
    this method will parse the URL into a 'noun' and a 'verb' :
 <PRE>noun: 'VacationAction'
verb: 'list'</PRE>

    It will then compare the noun-and-verb to the settings defined in <tt>web.xml</tt> (see {@link #init(ServletConfig)}). 
    If there is a match, then this method returns <tt>true</tt>.
  */
  public boolean usesUntrustedIdentifier(RequestParser aRequestParser) {
    boolean result = false; //by default, there is no ownership constraint
    String noun = extractNoun(aRequestParser); //WebUtil needs response too! servlet path + path info?
    if( isRestrictedRequest(noun) ){
      List<String> restrictedVerbs = fRestrictedOperations.get(noun);
      if ( hasAllOperationsRestricted(restrictedVerbs) ) {
        result = true;
      }
      else {
        String verb = extractVerb(aRequestParser);
        if ( restrictedVerbs.contains(verb) ) {
          result = true;
        }
      }
    }
    return result;
  }
  
  // PRIVATE 
  
  /** Holds all restricted operations in memory. Populated upon startup. */
  private static Map<String, List<String>> fRestrictedOperations = new LinkedHashMap<String, List<String>>();
  
  /** Name used in web.xml. */
  private static final String INIT_PARAM_NAME = "UntrustedProxyForUserId";
  
  /** Special character denoting all operations/verbs. */
  private static final String ALL_OPERATIONS = "*";
  
  private static final Logger fLogger = Util.getLogger(UntrustedProxyForUserIdImpl.class);
  
  private static void parseSettings(String aRawValue){
    fLogger.fine("Parsing ownership constraints defined in web.xml.");
    List<String> lines = parseSeparateLines(aRawValue);
    for(String line: lines){
       parseNounsAndVerbs(line);      
    }
  }
  
  private static List<String> parseSeparateLines(String aRawValue){
    List<String> result = new  ArrayList<String>();
    StringTokenizer parser = new StringTokenizer(aRawValue, "\n\r");
    while ( parser.hasMoreTokens() ) {
      result.add( parser.nextToken().trim() );
    }
    return result;    
  }

  private static void parseNounsAndVerbs(String aLine){
    String verb = WebUtil.getFileExtension(aLine);
    int verbStart = aLine.indexOf(".");
    String noun = aLine.substring(0,verbStart);
    if( isMissing(verb)  || isMissing(noun) ){
      throw new RuntimeException(
        "This line for the " + INIT_PARAM_NAME + " setting in web.xml does not have the expected form: " + Util.quote(aLine)
      );
    }
    add(noun.trim(), verb.trim());
  }
  
  private static boolean isMissing(String aText){
    return ! Util.textHasContent(aText);
  }
  
  private static void add(String aNoun, String aVerb){
    if( fRestrictedOperations.containsKey(aNoun) ) {
      addAnotherVerb(aNoun, aVerb);
    }
    else {
      addNewNounAndVerb(aNoun, aVerb);
    }
  }
  
  private static void addNewNounAndVerb(String aNoun, String aVerb){
    List<String> verbs = new ArrayList<String>();
    verbs.add(aVerb);
    fRestrictedOperations.put(aNoun, verbs);
  }
  
  private static void addAnotherVerb(String aNoun, String aVerb){
    if( ALL_OPERATIONS.equals(aVerb) ){
      fLogger.fine(Util.logOnePerLine(fRestrictedOperations));
      throw new RuntimeException(
        "When you use the '" + ALL_OPERATIONS + "' character to represent ALL operations, then only one line can be present for that item. " + 
        "In web.xml, you have a redundant setting for the init-param named " + INIT_PARAM_NAME + 
        " which needs to be removed. It is related to " +  Util.quote(aNoun) + " and " + Util.quote(aVerb)
      );
    }
    List<String> verbs = fRestrictedOperations.get(aNoun);
    verbs.add(aVerb);
  }

  private boolean isRestrictedRequest(String aNoun){
    return fRestrictedOperations.containsKey(aNoun);
  }
  
  private boolean hasAllOperationsRestricted(List<String> aVerbs){ 
    return aVerbs.contains(ALL_OPERATIONS);
  }
  
  /**
   For the example URL '.../BlahAction.do?X=Y', this method returns 'BlahAction' as the noun.
   Relies on the presence of '/' and '.' characters.
  */
  private String extractNoun(RequestParser aRequestParser){
    String uri = getURI(aRequestParser);
    int firstPeriod = uri.indexOf(".");
    if( firstPeriod == NOT_FOUND ) {
      throw new RuntimeException("Cannot find '.' character in URL: " + Util.quote(uri));
    }
    int lastSlash = uri.lastIndexOf("/");
    if( lastSlash == NOT_FOUND ) {
      throw new RuntimeException("Cannot find '/' character in URL: " + Util.quote(uri));
    }
    return uri.substring(lastSlash + 1, firstPeriod);
  }

  /** Return the part of the URL after the '.' character, but before any '?' character (if present). */
  private String extractVerb(RequestParser aRequestParser){
    return WebUtil.getFileExtension(getURI(aRequestParser));
  }
  
  private String getURI(RequestParser aRequestParser){
    return aRequestParser.getRequest().getRequestURI();
  }
}
